import 'package:evolution/evolution.dart';

///
/// Parallel Algorithm
///
/// An algorithm based on splitting the population
/// into one subpopulation subject to mutation and recombination
/// and another subpopulation of survivors that remain unchanged.

void parallelPopulations(
  int positions, // number of variables, i.e dimensionality of the problem
  int sizeN, // number of [Agent]s in the population
  int bestN, // number of [Agent]s selected by fitness
  int randN, // number of [Agent]s randomly selected
  int seed, // seeding the random number generator
  int steps, // number of generations
  double w, // weighting factor used in differential evolution
  fitness, // evaluation function
) {
  //int seed = DateTime.now().millisecond;
  var r = Random(seed);

  var z = 0;
  var p0 = generatePopulation(
    sizeN, // size of Population
    positions, // degrees of freedom
    r, // random number generator
    fitness, // fitness function
  );

  while (z < steps) {
    var wz = w / (z == 0 ? 1 : z).toDouble();

    // random survivors
    var p01 = p0.copy();
    p01.shuffle(r.r);
    var rand = p01.select(randN);

    // best survivors
    var p02 = p0.copy();
    var p021 = p02.sorted();
    var best = p021.select(bestN);

    // mutation
    var p03 = p0.copy();
    var mutated = p03.mutation(wz);

    //best crossover
    var p3 = best.copy();
    var p4 = p3.select(bestN);
    var bestCrossover = p4.crossover();

    //rand crossover
    var p6 = rand.copy();
    var p7 = p6.select(randN);
    var randCrossover = p7.crossover();

    // all
    var all = Population(
        /*rand +*/ best + mutated + bestCrossover + randCrossover,
        r,
        fitness);

    var p8 = all.copy();
    var p9 = p8.sorted();
    var p10 = p9.select(sizeN);

    p0 = p10;
    z++;
  }
  var res = p0.sorted().select(1);
  print(res);
}
